package sjson
package json

import dispatch.json._
trait Generic extends Protocol {

  import JsonSerialization._

  implicit def listFormat[T](implicit fmt : Format[T]) : Format[List[T]];

  def viaSeq[S <: Iterable[T], T] (f: Seq[T] => S) (implicit fmt : Format[T]) : Format[S] = new Format[S] {
    def writes(ts: S) = JsArray(ts.map(t => tojson(t)(fmt)).toList)
    def reads(json: JsValue) = json match {
      case JsArray(ts) => f(ts.map(t => fromjson[T](t)))
      case _ => throw new RuntimeException("Collection expected")
    }
  }

  /**
   * Use this when you would wrap a value in a case class
   *
   * <pre>
   * case class Name(name: String)
   * implicit val NameFormat: Format[Name] = wrap[Name, String]("name")(_.name, Name)
   *
   * val n = Name("debasish ghosh")
   * fromjson[Name](tojson(n)) should equal(n)
   * </pre>
   */
  def wrap[S, T](name: String)(to : S => T, from : T => S)(implicit fmt : Format[T]) = new Format[S]{
    def writes(s : S) = JsObject(List((tojson(name).asInstanceOf[JsString], tojson(to(s)))))
    def reads(js : JsValue) = js match {
      case JsObject(m) =>
        from(fromjson[T](m(JsString(name))))
      case _ => throw new RuntimeException("Object expected")
    }
  }

  /**
   * Lazy wrapper around serialization. Useful when you want to serialize mutually recursive structures.
   */
  def lazyFormat[S](fmt : => Format[S]) = new Format[S]{
    lazy val delegate = fmt;

    def reads(js : JsValue) = delegate.reads(js);
    def writes(s : S) = delegate.writes(s);
  }


  
  def asProduct2[S, T1,T2](f1: String,f2: String)(apply : (T1,T2) => S)(unapply : S => Product2[T1,T2])(implicit bin1: Format[T1],bin2: Format[T2]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct3[S, T1,T2,T3](f1: String,f2: String,f3: String)(apply : (T1,T2,T3) => S)(unapply : S => Product3[T1,T2,T3])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct4[S, T1,T2,T3,T4](f1: String,f2: String,f3: String,f4: String)(apply : (T1,T2,T3,T4) => S)(unapply : S => Product4[T1,T2,T3,T4])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct5[S, T1,T2,T3,T4,T5](f1: String,f2: String,f3: String,f4: String,f5: String)(apply : (T1,T2,T3,T4,T5) => S)(unapply : S => Product5[T1,T2,T3,T4,T5])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct6[S, T1,T2,T3,T4,T5,T6](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String)(apply : (T1,T2,T3,T4,T5,T6) => S)(unapply : S => Product6[T1,T2,T3,T4,T5,T6])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct7[S, T1,T2,T3,T4,T5,T6,T7](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String)(apply : (T1,T2,T3,T4,T5,T6,T7) => S)(unapply : S => Product7[T1,T2,T3,T4,T5,T6,T7])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct8[S, T1,T2,T3,T4,T5,T6,T7,T8](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8) => S)(unapply : S => Product8[T1,T2,T3,T4,T5,T6,T7,T8])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct9[S, T1,T2,T3,T4,T5,T6,T7,T8,T9](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9) => S)(unapply : S => Product9[T1,T2,T3,T4,T5,T6,T7,T8,T9])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct10[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10) => S)(unapply : S => Product10[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct11[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11) => S)(unapply : S => Product11[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct12[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12) => S)(unapply : S => Product12[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct13[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13) => S)(unapply : S => Product13[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct14[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14) => S)(unapply : S => Product14[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct15[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15) => S)(unapply : S => Product15[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct16[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16) => S)(unapply : S => Product16[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct17[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17) => S)(unapply : S => Product17[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct18[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String,f18: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18) => S)(unapply : S => Product18[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17],bin18: Format[T18]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17)),
          (tojson(f18).asInstanceOf[JsString], tojson(product._18))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17))),
          fromjson[T18](m(JsString(f18)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct19[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String,f18: String,f19: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19) => S)(unapply : S => Product19[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17],bin18: Format[T18],bin19: Format[T19]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17)),
          (tojson(f18).asInstanceOf[JsString], tojson(product._18)),
          (tojson(f19).asInstanceOf[JsString], tojson(product._19))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17))),
          fromjson[T18](m(JsString(f18))),
          fromjson[T19](m(JsString(f19)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct20[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String,f18: String,f19: String,f20: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20) => S)(unapply : S => Product20[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17],bin18: Format[T18],bin19: Format[T19],bin20: Format[T20]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17)),
          (tojson(f18).asInstanceOf[JsString], tojson(product._18)),
          (tojson(f19).asInstanceOf[JsString], tojson(product._19)),
          (tojson(f20).asInstanceOf[JsString], tojson(product._20))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17))),
          fromjson[T18](m(JsString(f18))),
          fromjson[T19](m(JsString(f19))),
          fromjson[T20](m(JsString(f20)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct21[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String,f18: String,f19: String,f20: String,f21: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21) => S)(unapply : S => Product21[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17],bin18: Format[T18],bin19: Format[T19],bin20: Format[T20],bin21: Format[T21]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17)),
          (tojson(f18).asInstanceOf[JsString], tojson(product._18)),
          (tojson(f19).asInstanceOf[JsString], tojson(product._19)),
          (tojson(f20).asInstanceOf[JsString], tojson(product._20)),
          (tojson(f21).asInstanceOf[JsString], tojson(product._21))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17))),
          fromjson[T18](m(JsString(f18))),
          fromjson[T19](m(JsString(f19))),
          fromjson[T20](m(JsString(f20))),
          fromjson[T21](m(JsString(f21)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  def asProduct22[S, T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22](f1: String,f2: String,f3: String,f4: String,f5: String,f6: String,f7: String,f8: String,f9: String,f10: String,f11: String,f12: String,f13: String,f14: String,f15: String,f16: String,f17: String,f18: String,f19: String,f20: String,f21: String,f22: String)(apply : (T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22) => S)(unapply : S => Product22[T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22])(implicit bin1: Format[T1],bin2: Format[T2],bin3: Format[T3],bin4: Format[T4],bin5: Format[T5],bin6: Format[T6],bin7: Format[T7],bin8: Format[T8],bin9: Format[T9],bin10: Format[T10],bin11: Format[T11],bin12: Format[T12],bin13: Format[T13],bin14: Format[T14],bin15: Format[T15],bin16: Format[T16],bin17: Format[T17],bin18: Format[T18],bin19: Format[T19],bin20: Format[T20],bin21: Format[T21],bin22: Format[T22]) = new Format[S]{
    def writes(s: S) = {
      val product = unapply(s)
      JsObject(
        List(
          (tojson(f1).asInstanceOf[JsString], tojson(product._1)),
          (tojson(f2).asInstanceOf[JsString], tojson(product._2)),
          (tojson(f3).asInstanceOf[JsString], tojson(product._3)),
          (tojson(f4).asInstanceOf[JsString], tojson(product._4)),
          (tojson(f5).asInstanceOf[JsString], tojson(product._5)),
          (tojson(f6).asInstanceOf[JsString], tojson(product._6)),
          (tojson(f7).asInstanceOf[JsString], tojson(product._7)),
          (tojson(f8).asInstanceOf[JsString], tojson(product._8)),
          (tojson(f9).asInstanceOf[JsString], tojson(product._9)),
          (tojson(f10).asInstanceOf[JsString], tojson(product._10)),
          (tojson(f11).asInstanceOf[JsString], tojson(product._11)),
          (tojson(f12).asInstanceOf[JsString], tojson(product._12)),
          (tojson(f13).asInstanceOf[JsString], tojson(product._13)),
          (tojson(f14).asInstanceOf[JsString], tojson(product._14)),
          (tojson(f15).asInstanceOf[JsString], tojson(product._15)),
          (tojson(f16).asInstanceOf[JsString], tojson(product._16)),
          (tojson(f17).asInstanceOf[JsString], tojson(product._17)),
          (tojson(f18).asInstanceOf[JsString], tojson(product._18)),
          (tojson(f19).asInstanceOf[JsString], tojson(product._19)),
          (tojson(f20).asInstanceOf[JsString], tojson(product._20)),
          (tojson(f21).asInstanceOf[JsString], tojson(product._21)),
          (tojson(f22).asInstanceOf[JsString], tojson(product._22))
        ))
    }
    def reads(js: JsValue) = js match {
      case JsObject(m) => // m is the Map
        apply(
          fromjson[T1](m(JsString(f1))),
          fromjson[T2](m(JsString(f2))),
          fromjson[T3](m(JsString(f3))),
          fromjson[T4](m(JsString(f4))),
          fromjson[T5](m(JsString(f5))),
          fromjson[T6](m(JsString(f6))),
          fromjson[T7](m(JsString(f7))),
          fromjson[T8](m(JsString(f8))),
          fromjson[T9](m(JsString(f9))),
          fromjson[T10](m(JsString(f10))),
          fromjson[T11](m(JsString(f11))),
          fromjson[T12](m(JsString(f12))),
          fromjson[T13](m(JsString(f13))),
          fromjson[T14](m(JsString(f14))),
          fromjson[T15](m(JsString(f15))),
          fromjson[T16](m(JsString(f16))),
          fromjson[T17](m(JsString(f17))),
          fromjson[T18](m(JsString(f18))),
          fromjson[T19](m(JsString(f19))),
          fromjson[T20](m(JsString(f20))),
          fromjson[T21](m(JsString(f21))),
          fromjson[T22](m(JsString(f22)))
        )
      case _ => throw new RuntimeException("object expected")
    }
  }  

  /**
   * one sample for i=3 that will be generated is:
   * def asProduct3[S, T1, T2, T3](f1: String, f2: String, f3: String)(apply : (T1, T2, T3) => S)(unapply : S => Product3[T1, T2, T3])(implicit bin1: Format[T1], bin2: Format[T2], bin3: Format[T3]) = new Format[S]{
   *   def writes(s: S) = {
   *     val product = unapply(s)
   *     JsObject(
   *       List(
   *         (tojson(f1).asInstanceOf[JsString], tojson(product._1)), 
   *         (tojson(f2).asInstanceOf[JsString], tojson(product._2)), 
   *         (tojson(f3).asInstanceOf[JsString], tojson(product._3)) 
   *     ))
   *   }
   *   def reads(js: JsValue) = js match {
   *     case JsObject(m) => // m is the Map
   *       apply(
   *         fromjson[T1](m(JsString(f1))), 
   *         fromjson[T2](m(JsString(f2))), 
   *         fromjson[T3](m(JsString(f3)))
   *       )
   *     case _ => throw new RuntimeException("object expected")
   *   }
   * }  
   **/

  /**
  case class Summand[T](clazz : Class[_], format : Format[T]);
  implicit def classToSummand[T](clazz : Class[T])(implicit fmt : Format[T]) : Summand[T] = Summand[T](clazz, fmt);
  implicit def formatToSummand[T](format : Format[T])(implicit mf : scala.reflect.Manifest[T]) : Summand[T] = 
    Summand[T](mf.erasure, format);

  def asUnion[S](summands : Summand[_ <: S]*) : Format[S] = 
    if (summands.length >= 256) error("Sums of 256 or more elements currently not supported");
    else
    new Format[S]{
      val mappings = summands.toArray.zipWithIndex;

      def reads(json : JsValue) : S = read(in)(summands(read[Byte](in)).format)
      // def writes(ts: S) = JsArray(ts.map(t => tojson(t)(fmt)).toList)
      // def reads(json: JsValue) = json match {

      def writes(ts: S) = JsArray(ts.map(t => tojson(t)(fmt)).toList)

      def writes(out : Output, s : S): Unit =
        mappings.find(_._1.clazz.isInstance(s)) match {
          case Some( (sum, i) ) => writeSum(out, s, sum, i)
          case None => error("No known sum type for object " + s);
        }
      private def writeSum[T](out : Output, s : S, sum : Summand[T], i : Int) {
        write(out, i.toByte);
        // 2.7/2.8 compatibility: cast added by MH
        write(out, sum.clazz.cast(s).asInstanceOf[T])(sum.format);
      }
  }
**/
}
